@@ -2,10 +2,25 @@ from struct import pack, unpack
 from uuid import UUID
 from datetime import datetime
 from enum import Enum
+from typing import Any, TypeVar
 
+# constants
 ticksBetweenEpochs = 621355968000000000
 dateMask = 0x3fffffffffffffff
 
+class UnionDefinition:
+    """
+    A utility class for unions
+    """
+    discriminator: int
+    value: Any
+
+    def __init__(self, discriminator: int, value: Any):
+        self.discriminator = discriminator
+        self.value = value
+
+UnionType = TypeVar("UnionType", bound=UnionDefinition)
+
 class BebopReader:
     """
     A wrapper around a bytearray for reading Bebop base types from it.
@@ -88,23 +103,22 @@ class BebopReader:
         length = self.read_uint32()
         if length == 0:
             return self._emptyString
-        v = self._buffer[self.index : self.index + length]
+        string_data = self._buffer[self.index : self.index + length]
         self.index += length
-        return "".join([chr(c) for c in v])
+        return string_data.decode('utf-8')
 
     def read_guid(self) -> UUID:
-        g = UUID(bytes_le=bytes(self._buffer[self.index : self.index + 16]))
+        b = self._buffer[self.index : self.index + 16]
+        reordered = [b[3], b[2], b[1], b[0], b[5], b[4], b[7], b[6], b[8], b[9], b[10], b[11], b[12], b[13], b[14], b[15]]
+        g = UUID(bytes=bytes(reordered))
         self.index += 16
         return g
 
     def read_date(self) -> datetime:
         ticks = self.read_uint64() & dateMask
-        ms = (ticks - ticksBetweenEpochs)
+        ms = (ticks - ticksBetweenEpochs) / 10000000
         return datetime.fromtimestamp(ms)
 
-    def read_enum(self, values: list):
-        return values[self.read_uint32()]
-
     read_message_length = read_uint32
 
 
@@ -122,11 +136,10 @@ class BebopWriter:
 
     def _guarantee_buffer_length(self):
         """
-        Replace _buffer with a bytearray twice the size required to ensure buffer is always big enough
+        This is only needed when message length is unknown; only occurs when _grow_by is used 
         """
         if self.length > len(self._buffer):
-            new_length = min(2 * len(self._buffer), self.length)
-            data = bytearray([0] * new_length)
+            data = bytearray([0] * self.length)
             data[:len(self._buffer)] = self._buffer
             self._buffer = data
 
@@ -186,19 +199,18 @@ class BebopWriter:
         if len(val) == 0:
             self.write_uint32(0)
             return
-        self.write_bytes(val.encode())
+        self.write_bytes(val.encode("utf-8"))
 
     def write_guid(self, guid: UUID):
-        self.write_bytes(guid.bytes_le, write_msg_length=False)
+        b = guid.bytes
+        bebop_uid = [b[3], b[2], b[1], b[0], b[5], b[4], b[7], b[6], b[8], b[9], b[10], b[11], b[12], b[13], b[14], b[15]]
+        self.write_bytes(bebop_uid, write_msg_length=False)
 
     def write_date(self, date: datetime):
-        ms = int(datetime.timestamp(date))
-        ticks = ms + ticksBetweenEpochs
+        ms = int(date.timestamp())
+        ticks = ms * 10000000 + ticksBetweenEpochs 
         self.write_uint64(ticks & dateMask)
 
-    def write_enum(self, val: Enum):
-        self.write_uint32(val.value)
-
     def reserve_message_length(self):
         """
         Reserve some space to write a message's length prefix, and return its index.
